/*********************************************************************

  ui_text.c

  Functions used to retrieve text used by MAME, to aid in
  translation.

*********************************************************************/

#include "driver.h"
#include "ui_text.h"

char portname[PORT_last_entry + 1][MAX_PORT_LEN + 1];

/* All entries in this table must match the enum ordering in "ui_text.h" */
#ifdef BUILD_ENGLISH
const char * default_text[] =
{
	"CANAME",

	/* copyright stuff */
	"Usage of emulators in conjunction with ROMs you don't own is forbidden by copyright law.",
	"IF YOU ARE NOT LEGALLY ENTITLED TO PLAY \"%s\" ON THIS EMULATOR, PRESS ESC.",
	"Otherwise, type OK to continue",

	/* misc stuff */
	"Return to Main Menu",
	"Return to Prior Menu",
	"Press Any Key",
	"Press Select Key/Button",
	"On",
	"Off",
	"NA",
	"OK",
	"INVALID",
	"(none)",
	"CPU",
	"Address",
	"Value",
	"Sound",
	"sound",
	"stereo",
	"Vector Game",
	"Screen Resolution",
	"Text",
	"Volume",
	"Relative",
	"ALL CHANNELS",
	"Brightness",
	"Gamma",
	"Vector Flicker",
	"Vector Intensity",
	"Overclock",
	"ALL CPUS",
	"History not available",
	"Command List not available",

	/* special characters */
	"\x11",
	"\x10",
	"\x18",
	"\x19",
	"\x1a",
	"\x1b",

	/* known problems */
	"There are known problems with this game:",
	"The colors aren't 100% accurate.",
	"The colors are completely wrong.",
	"The video emulation isn't 100% accurate.",
	"The sound emulation isn't 100% accurate.",
	"The game lacks sound.",
	"Screen flipping in cocktail mode is not supported.",
	"THIS GAME DOESN'T WORK PROPERLY",
	"The game has protection which isn't fully emulated.",
	"There are working clones of this game. They are:",
	"Type OK to continue",

	/* main menu */
	"Input (general)",
	"Dip Switches",
	"Analog Controls",
	"Bookkeeping Info",
	"Input (this game)",
	"Game Information",
	"Game History",
	"Show Command List",
	"Custom Buttons",
	"Autofire Setting",
	"Reset Game",
	"Return to Game",
	"Cheat",
	"Memory Card",

	/* input */
	"Key/Joy Speed",
	"Reverse",
	"Sensitivity",

	/* stats */
	"Tickets dispensed",
	"Coin",
	"(locked)",

	/* memory card */
	"Load Memory Card",
	"Eject Memory Card",
	"Create Memory Card",
	"Call Memory Card Manager (RESET)",
	"Failed To Load Memory Card!",
	"Load OK!",
	"Memory Card Ejected!",
	"Memory Card Created OK!",
	"Failed To Create Memory Card!",
	"(It already exists ?)",
	"DAMN!! Internal Error!",

	/* cheats */
	"Enable/Disable a Cheat",
	"Add/Edit a Cheat",
	"Start a New Cheat Search",
	"Continue Search",
	"View Last Results",
	"Restore Previous Results",
	"Configure Watchpoints",
	"General Help",
	"Options",
	"Reload Database",
	"Watchpoint",
	"Disabled",
	"Cheats",
	"Watchpoints",
	"More Info",
	"More Info for",
	"Name",
	"Description",
	"Activation Key",
	"Code",
	"Max",
	"Set",
	"Cheat conflict found: disabling",
	"Help not available yet",

	/* watchpoints */
	"Number of bytes",
	"Display Type",
	"Label Type",
	"Label",
	"X Position",
	"Y Position",
	"Watch",

	"Hex",
	"Decimal",
	"Binary",

	/* searching */
	"Lives (or another value)",
	"Timers (+/- some value)",
	"Energy (greater or less)",
	"Status (bits or flags)",
	"Slow But Sure (changed or not)",
	"Default Search Speed",
	"Fast",
	"Medium",
	"Slow",
	"Very Slow",
	"All Memory",
	"Select Memory Areas",
	"Matches found",
	"Search not initialized",
	"No previous values saved",
	"Previous values already restored",
	"Restoration successful",
	"Select a value",
	"All values saved",
	"One match found - added to list",

	/* custom button */
	"Custom Button ",
	"Not Use",

	/* autofire */
	"   OFF",
	"    ON",
	"TOGGLE",
	"Autofire Delay",
	"Autofire %s",
	
	/* Wave recording */
	"WAVE Recording...",

	/* Juke Box */
	"- JUKE BOX -",
	"Bank",
	"Code 1",
	"Code 2",
	"COMMAND",
	"PLAY",
	"STOP",
	"RECORD",
	"STOP  : S",
	"RECORD: R",
	"CODE  : C",
	"NOTES : N",
	"FF(\x10\x10): F",
	"Fast-Forwarding",
	"Go to Play List Menu",
	"Go to Manual Code Menu",
	"BGM",
	"SE1",
	"SE2",
	"SE3",
	"SE4",
	"SE5",
	"SE6",
	"SE7",

	/* object operation */
	"- OBJECT OPERATION -",
	"Object Number",
	"Object Output",
	"Background Number",
	"Background Output",
	"Front Layer",
	"Alt Backcolor",
	"All Output Clear",
	"Set Default",
	"OFF",
	" ON",

	/* input.c */
	"None",
	"not",
	"or",
	"n/a",

	/* window.c */
	"Full Screen\tAlt+Enter",

	/* neogeo.c */
	"raster effects %s",
	"enabled",
	"disabled",

	NULL
};

const char *default_port[] =
{
	"Config Menu",
	"On Screen Display",
	"Pause",
	"Single Step",
	"Reset Game",
	"Show Gfx",
	"Frameskip Dec",
	"Frameskip Inc",
	"Throttle",
	"Show FPS",
	"Show Profiler",
	"Save Snapshot",
	"Toggle Cheat",
	"Save State",
	"Load State",
	"Add Cheat",
	"Delete Cheat",
	"Save Cheat",
	"Watch Value",
	"Change Screen Mode",
	"Toggle Autofire",
	"Toggle Wave Recording",
	"Show Input Log",
	"Object Operation",
	"Show Command List",
	"Show Cheat List",
	"Show the Time",
	"UI Direction\x81\xd8",
	"UI Direction\x81\xd2",
	"UI Direction\x81\xd4",
	"UI Direction\x81\xd6",
	"UI Select",
	"UI Cancel",
	"Emulation Close",
	"Coin 1",
	"Coin 2",
	"Coin 3",
	"Coin 4",
	"Service",
	"P1 Start",
	"P1 Select",
	"P1 Direction\x81\xd8",
	"P1 Direction\x81\xd2",
	"P1 Direction\x81\xd4",
	"P1 Direction\x81\xd6",
	"P1 Button\x81\xc0",
	"P1 Button\x81\xc1",
	"P1 Button\x81\xc2",
	"P1 Button\x81\xc3",
	"P1 Button\x81\xe9",
	"P1 Button\x81\xea",
	"P1 Button\x81\xeb",
	"P1 Button\x81\xec",
	"P1 Button\x81\xed",
	"P1 Button\x81\xee",
	"P1 Custom Button 1",
	"P1 Custom Button 2",
	"P1 Custom Button 3",
	"P1 Custom Button 4",
	"P1 Track Y - (\x81\xd8)",
	"P1 Track Y + (\x81\xd2)",
	"P1 Track X - (\x81\xd4)",
	"P1 Track X + (\x81\xd6)",
	"P1 Dial -",
	"P1 Dial +",
	"P2 Start",
	"P2 Select",
	"P2 Direction\x81\xd8",
	"P2 Direction\x81\xd2",
	"P2 Direction\x81\xd4",
	"P2 Direction\x81\xd6",
	"P2 Button\x81\xc0",
	"P2 Button\x81\xc1",
	"P2 Button\x81\xc2",
	"P2 Button\x81\xc3",
	"P2 Button\x81\xe9",
	"P2 Button\x81\xea",
	"P2 Button\x81\xeb",
	"P2 Button\x81\xec",
	"P2 Button\x81\xed",
	"P2 Button\x81\xee",
	"P2 Custom Button 1",
	"P2 Custom Button 2",
	"P2 Custom Button 3",
	"P2 Custom Button 4",
	"P2 Track Y - (\x81\xd8)",
	"P2 Track Y + (\x81\xd2)",
	"P2 Track X - (\x81\xd4)",
	"P2 Track X + (\x81\xd6)",
	"P2 Dial -",
	"P2 Dial +",
	"P3 Start",
	"P3 Direction\x81\xd8",
	"P3 Direction\x81\xd2",
	"P3 Direction\x81\xd4",
	"P3 Direction\x81\xd6",
	"P3 Button\x81\xe9",
	"P3 Button\x81\xea",
	"P3 Button\x81\xeb",
	"P3 Button\x81\xec",
	"P4 Start",
	"P4 Direction\x81\xd8",
	"P4 Direction\x81\xd2",
	"P4 Direction\x81\xd4",
	"P4 Direction\x81\xd6",
	"P4 Button\x81\xe9",
	"P4 Button\x81\xea",
	"P4 Button\x81\xeb",
	"P4 Button\x81\xec",
	"UnKnown",

	NULL
};

char ipdn_defaultstrings[][MAX_DEFSTR_LEN] =
{
	"Off",
	"On",
	"No",
	"Yes",
	"Lives",
	"Bonus Life",
	"Difficulty",
	"Demo Sounds",
	"Coinage",
	"Coin A",
	"Coin B",
	"9 Coins/1 Credit",
	"8 Coins/1 Credit",
	"7 Coins/1 Credit",
	"6 Coins/1 Credit",
	"5 Coins/1 Credit",
	"4 Coins/1 Credit",
	"3 Coins/1 Credit",
	"8 Coins/3 Credits",
	"4 Coins/2 Credits",
	"2 Coins/1 Credit",
	"5 Coins/3 Credits",
	"3 Coins/2 Credits",
	"4 Coins/3 Credits",
	"4 Coins/4 Credits",
	"3 Coins/3 Credits",
	"2 Coins/2 Credits",
	"1 Coin/1 Credit",
	"4 Coins/5 Credits",
	"3 Coins/4 Credits",
	"2 Coins/3 Credits",
	"4 Coins/7 Credits",
	"2 Coins/4 Credits",
	"1 Coin/2 Credits",
	"2 Coins/5 Credits",
	"2 Coins/6 Credits",
	"1 Coin/3 Credits",
	"2 Coins/7 Credits",
	"2 Coins/8 Credits",
	"1 Coin/4 Credits",
	"1 Coin/5 Credits",
	"1 Coin/6 Credits",
	"1 Coin/7 Credits",
	"1 Coin/8 Credits",
	"1 Coin/9 Credits",
	"Free Play",
	"Cabinet",
	"Upright",
	"Cocktail",
	"Flip Screen",
	"Service Mode",
	"Test Switch",
	"Coin Chutes?",
	"Autofire (in some games)",
	"COMM Setting",
	"Freeze",
	"Territory",
	"Machine Mode",
	"Japan",
	"USA",
	"Europe",
	"Home",
	"Arcade",
	"Allow Continue",
	"Unused",
	"Unknown"
};
#else
const char * default_text[] =
{
	"CANAME",

	/* copyright stuff */
	"LĂȂROMG~[^Ŏgp邱Ƃ́A쌠@ɂċւĂ܂B",
	"\"%s\" @IɗVԌȂESCĂB",
	"OKƓ͂Ƒ܂B",

	/* misc stuff */
	"Cj[ɖ߂",
	"Õj[ɖ߂",
	"L[Ă",
	"L[/{^Ă",
	"ON",
	"OFF",
	"Ȃ",
	"OK",
	"",
	"(Ȃ)",
	"CPU",
	"AhX",
	"l",
	"",
	"TEhp",
	"XeI",
	"xN^Q[",
	"ʉ𑜓x",
	"eLXg",
	"",
	"",
	"S`l",
	"邳",
	"K}l",
	"xN^tbJ[",
	"xN^x",
	"I[o[NbN",
	"SCPU",
	"qXg[܂",
	"R}hXg܂",

	/* special characters */
	"\x11",
	"\x10",
	"\x18",
	"\x19",
	"\x1a",
	"\x1b",

	/* known problems */
	"̃Q[ɂ͈ȉ̖肪܂:",
	"F̍ČsSłB",
	"FSɊԈĂ܂B",
	"rfIG~[VsSłB",
	"TEhG~[VsSłB",
	"TEhĐ܂B",
	"JNe[h̉ʔ]ɑΉĂ܂B",
	"̃Q[͑S삵܂",
	"̃Q[̓veNĝ߃G~[VsSłB",
	"̃Q[ɂ͈ȉ̓\ȃN[݂܂:",
	"OKƓ͂Ƒ܂",

	/* main menu */
	"͐ݒ (W)",
	"fBbvXCb`",
	"AiORg[",
	"CJ",
	"͐ݒ (̃Q[)",
	"Q[",
	"Q[̗j",
	"R}hꗗ/",
	"JX^{^ݒ",
	"Aːݒ",
	"Q[Zbg",
	"Q[ɖ߂",
	"`[g (Q[̉)",
	"J[h",

	/* input */
	"Key/Joy x",
	"]",
	"x",

	/* stats */
	"`Pbg",
	"RC",
	"(Œ)",

	/* memory card */
	"J[h̓ǂ݂",
	"J[h̎o",
	"J[h̍쐬",
	"J[h}l[WĂяo (Zbg)",
	"J[h̓ǂ݂ݎs!",
	"ǂ݂݊!",
	"J[ho܂!",
	"J[h쐬܂!",
	"J[h̍쐬s!",
	"(łɍ쐬ς?)",
	"G[ł!!",

	/* cheats */
	"`[gL/ɂ",
	"`[gǉ/ҏW",
	"VK`[g̃T[`Jn",
	"T[`p",
	"Ō̌ʂ\",
	"Ǒʂɖ߂",
	"EHb`|Cg̐ݒ",
	"Sʂ̃wv",
	"IvV",
	"f[^x[Xēǂݍ",
	"EHb`|Cg",
	"",
	"`[g",
	"EHb`|Cg",
	"⑫",
	"⑫:",
	"O",
	"",
	"LL[",
	"R[h",
	"ő",
	"Zbg",
	"d`[g܂: ɂ܂",
	"wv͂܂܂",

	/* watchpoints */
	"oCg",
	"\`",
	"x^Cv",
	"x",
	"XW",
	"YW",
	"EHb`",

	"16i",
	"10i",
	"2i",

	/* searching */
	"Ct (܂l)",
	"^C}[ (+/- ̒l)",
	"GlM[ (l̑召)",
	" (rbgtO)",
	"xǊm (ωȂ)",
	"W̃T[`x",
	"",
	"",
	"x",
	"ƂĂx",
	"SẴ",
	"̈I",
	"vl܂",
	"T[`Ă܂",
	"O̒lۑĂ܂",
	"O̒l͊ɕAς݂ł",
	"l̕Aɐ܂",
	"l̑I",
	"ׂĂ̒lۑ܂",
	"vl܂ - Xgɒǉ܂",

	/* custom button */
	"JX^{^",
	"gpȂ",

	/* autofire */
	"    It",
	"펞I",
	"؂ւ",
	"AˊԊu",
	"Aː؂ւ %s",
	
	/* Wave recording */
	"^",

	/* Juke Box */
	"- W[N{bNX -",
	"oN",
	"R[h1",
	"R[h2",
	"R}h",
	"tJn",
	"t~",
	"^Jn",
	"t~  : S",
	"^Jn  : R",
	"R[h\: C",
	"ߕ\  : N",
	"(\x10\x10): F",
	"蒆",
	"TEhXgʂ֍s",
	"R[h͉ʂ֍s",
	"obNTEh",
	"Eʉ 1",
	"Eʉ 2",
	"Eʉ 3",
	"Eʉ 4",
	"Eʉ 5",
	"Eʉ 6",
	"Eʉ 7",

	/* object operation */
	"- IuWFNg -",
	"IuWFNgԍ",
	"IuWFNgo",
	"obNOEhԍ",
	"obNOEho",
	"tgCA[",
	"wiF̕ύX",
	"So͂NA",
	"ftHgɖ߂",
	"It",
	"I",

	/* input.c */
	"gp",
	"",
	"",
	"gps",

	/* window.c */
	"tXN[\tAlt+Enter",

	/* neogeo.c */
	"X^GtFNg %s",
	"L",
	"",

	NULL
};

const char *default_port[] =
{
	"ݒ胁j[",
	"IXN[fBXvC",
	"|[Y",
	"R}",
	"Q[̃Zbg",
	"OtBbNf[^\",
	"t[XLbv",
	"t[XLbv",
	"Xbg",
	"FPS\",
	"vt@C\",
	"XibvVbgۑ",
	"`[g؂ւ",
	"vCԂ̕ۑ",
	"vCԂ̕A",
	"`[gǉ",
	"`[g폜",
	"`[gۑ",
	"EHb`̒l\",
	"fBXvC[h؂ւ",
	"Aː؂ւ",
	"WAV^Jn/I",
	"̓R}h̕\",
	"IuWFNg",
	"R}h/@̕\",
	"`[gXg̕\",
	"̕\",
	"UI \x81\xd8ړ",
	"UI \x81\xd2ړ",
	"UI \x81\xd4ړ",
	"UI \x81\xd6ړ",
	"UI I",
	"UI LZ",
	"G~[VI",
	"RC1",
	"RC2",
	"RC3",
	"RC4",
	"T[rX",
	"P1 X^[g",
	"P1 ZNg",
	"P1 \x81\xd8",
	"P1 \x81\xd2",
	"P1 \x81\xd4",
	"P1 \x81\xd6",
	"P1 {^\x81\xc0",
	"P1 {^\x81\xc1",
	"P1 {^\x81\xc2",
	"P1 {^\x81\xc3",
	"P1 {^\x81\xe9",
	"P1 {^\x81\xea",
	"P1 {^\x81\xeb",
	"P1 {^\x81\xec",
	"P1 {^\x81\xed",
	"P1 {^\x81\xee",
	"P1 JX^{^1",
	"P1 JX^{^2",
	"P1 JX^{^3",
	"P1 JX^{^4",
	"P1 gbN{[ \x81\xd8ړ",
	"P1 gbN{[ \x81\xd2ړ",
	"P1 gbN{[ \x81\xd4ړ",
	"P1 gbN{[ \x81\xd6ړ",
	"P1 _C|",
	"P1 _C{",
	"P2 X^[g",
	"P2 ZNg",
	"P2 \x81\xd8",
	"P2 \x81\xd2",
	"P2 \x81\xd4",
	"P2 \x81\xd6",
	"P2 {^\x81\xc0",
	"P2 {^\x81\xc1",
	"P2 {^\x81\xc2",
	"P2 {^\x81\xc3",
	"P2 {^\x81\xe9",
	"P2 {^\x81\xea",
	"P2 {^\x81\xeb",
	"P2 {^\x81\xec",
	"P2 {^\x81\xed",
	"P2 {^\x81\xee",
	"P2 JX^{^1",
	"P2 JX^{^2",
	"P2 JX^{^3",
	"P2 JX^{^4",
	"P2 gbN{[ \x81\xd8ړ",
	"P2 gbN{[ \x81\xd2ړ",
	"P2 gbN{[ \x81\xd4ړ",
	"P2 gbN{[ \x81\xd6ړ",
	"P2 _C|",
	"P2 _C{",
	"P3 X^[g",
	"P3 \x81\xd8",
	"P3 \x81\xd2",
	"P3 \x81\xd4",
	"P3 \x81\xd6",
	"P3 {^\x81\xe9",
	"P3 {^\x81\xea",
	"P3 {^\x81\xeb",
	"P3 {^\x81\xec",
	"P4 X^[g",
	"P4 \x81\xd8",
	"P4 \x81\xd2",
	"P4 \x81\xd4",
	"P4 \x81\xd6",
	"P4 {^\x81\xe9",
	"P4 {^\x81\xea",
	"P4 {^\x81\xeb",
	"P4 {^\x81\xec",
	"s",

	NULL
};

char ipdn_defaultstrings[][MAX_DEFSTR_LEN] =
{
	"OFF",
	"ON",
	"",
	"͂",
	"Ct",
	"{[iXCt",
	"Փx",
	"fTEh",
	"RC",
	"RCA",
	"RCB",
	"9RC/1NWbg",
	"8RC/1NWbg",
	"7RC/1NWbg",
	"6RC/1NWbg",
	"5RC/1NWbg",
	"4RC/1NWbg",
	"3RC/1NWbg",
	"8RC/3NWbg",
	"4RC/2NWbg",
	"2RC/1NWbg",
	"5RC/3NWbg",
	"3RC/2NWbg",
	"4RC/3NWbg",
	"4RC/4NWbg",
	"3RC/3NWbg",
	"2RC/2NWbg",
	"1RC/1NWbg",
	"4RC/5NWbg",
	"3RC/4NWbg",
	"2RC/3NWbg",
	"4RC/7NWbg",
	"2RC/4NWbg",
	"1RC/2NWbg",
	"2RC/5NWbg",
	"2RC/6NWbg",
	"1RC/3NWbg",
	"2RC/7NWbg",
	"2RC/8NWbg",
	"1RC/4NWbg",
	"1RC/5NWbg",
	"1RC/6NWbg",
	"1RC/7NWbg",
	"1RC/8NWbg",
	"1RC/9NWbg",
	"t[vC",
	"➑",
	"AbvCg",
	"JNe",
	"ʔ]",
	"T[rX[h",
	"eXgXCb`",
	"RCH",
	"A (̃Q[ŗL)",
	"ʐMݒ",
	"t[Y",
	"n",
	"}V[h",
	"{",
	"AJ",
	"[bp",
	"ƒp",
	"A[P[h",
	"ReBj[\",
	"gp",
	"s"
};
#endif

const char * ui_getstring (int string_num)
{
	return default_text[string_num];
}

void portname_init(void)
{
	int i;

	for (i = PORT_first_entry + 1; i < PORT_last_entry; i++)
		strcpy(portname[i], default_port[i]);
}
